(define (ast-type x)
  (define (builtin? x)
    (case x
      ('+ #t)
      ('- #t)
      ('* #t)
      ('! #t)
      ('= #t)
      ('bool->int #t)
      ('print #t)
      (else #f)))
  (cond
   ((list? x)
    (case (car x)
      ('if 'if)
      ('let 'let)
      ('lambda 'lambda)
      ('closure 'closure) ; only available in codegen
      ('static-string 'static-string) ; only available in codegen
      (else 'app)))
   ((builtin? x) 'builtin)
   ((symbol? x) 'var)
   ((integer? x) 'int-literal)
   ((boolean? x) 'bool-literal)
   ((string? x) 'string-literal)))

(define (ast-traverse f x)
  (case (ast-type x)
    ('let `(let ,(map (lambda (x) (list (car x) (f (cadr x))))
		      (let-bindings x))
	     ,@(map f (let-body x))))
    ('app (map f x))
    ('lambda `(lambda ,(lambda-args x) ,(f (lambda-body x))))
    ('if `(if ,@(map f (cdr x))))
    (else x)))

(define (ast-collect f x)
  (define (inner y) (ast-collect f y))
  (case (ast-type x)
    ['let (append (f x)
		  (flat-map inner (let-bindings x))
		  (flat-map inner (let-body x)))]
    ['app (append (f x)
		  (flat-map inner x))]
    ['lambda (append (f x)
		     (inner (lambda-body x)))]
    ['if (append (f x)
		 (flat-map inner (cdr x)))]
    ['closure (flat-map inner (caddr x))]
    [else (f x)]))

(define (ast-find p x)
  (define (inner y) (ast-find p y))
  (define (any p x) (fold-left
		     (lambda (acc y) (if acc #t (p y)))
		     #f
		     x))
  (define (either . fs)
    (if (null? fs) #f
	(if (car fs) (car fs)
	    (apply either (cdr fs)))))
		     
  (case (ast-type x)
    ['let (either (p x)
		  (any inner (let-bindings x))
		  (any inner (let-body x)))]
    ['app (either (p x)
		  (any inner x))]
    ['lambda (either (p x)
		     (inner (lambda-body x)))]
    ['if (either (p x) (any inner (cdr x)))]
    [else (p x)]))

(define (let-bindings e)
  (define (pattern-match binding body)
    (if (eqv? (ast-type binding) 'var)
	(list (cons binding body))
	(let* ([constructor (car binding)]
	       [destructor (lambda (i) (dtor-name constructor i))])
	  (flat-map (lambda (y i)
		      (pattern-match y `((,(destructor i) ,@body))))
		    (cdr binding)
		    (range 0 (length (cdr binding)))))))
  (flat-map (lambda (x) (pattern-match (car x) (cdr x))) (cadr e)))
(define let-body cddr)

(define (lambda? x)
  (and (list? x) (eq? (car x) 'lambda)))


(define (statement-type x)
  (cond
   [(and (list? x)
	 (eqv? (car x) 'data)) 'data]
   [(and (list? x)
	 (eqv? (car x) 'define)) 'define]
   [else 'expr]))

(define (program-datas program)
  (filter (lambda (x) (eqv? (statement-type x) 'data))
	  program))

(define (program-defines program)
  (filter (lambda (x) (eqv? (statement-type x) 'defines))
	  program))

(define (program-body program)
  ; hack to have multi-expression bodies
  `(let ()
     ,@(filter (lambda (x) (eqv? (statement-type x) 'expr))
	       program)))


					; (A ((foo (Int Bool))
					;     (bar (Bool)))

(define data-layout cdr)

					; gets both constructors and destructors
					; (data A (foo Int Bool)
					;         (bar Bool))
					;        |
					;        v
					; (foo . (constructor . (abs Int (abs Bool A))))
					; (foo~0 . (0 . (abs A Int)))
					; (foo~1 . (1 . (abs A Bool)))
					; (bar . (constructor . (abs Bool A)))
					; (bar~0 . (0 . (abs A Bool)))

(define (data-tors data-layout)
  (define (constructor-type t products)
    (fold-right (lambda (x acc) `(abs ,x ,acc)) t products))

  (define (destructor ctor-name prod-type part-type index)
    (let ([name (dtor-name ctor-name index)])
      (cons name (cons index `(abs ,prod-type ,part-type)))))
  
  (let ([type-name (car data-layout)]
        [ctors (cdr data-layout)])
    (fold-right
     (lambda (ctor acc)       
       (let* ([ctor-name (car ctor)]
	      [products (cdr ctor)]
	      
	      [maker (cons ctor-name (cons 'constructor (constructor-type type-name products)))]
	      
	      [dtors (map (lambda (t i) (destructor ctor-name type-name t i))
			  products
			  (range 0 (length products)))])
	 
	 (cons maker (append dtors acc))))
     '()
     ctors)))

					; creates a type environment for a given adt definition
(define (data-tors-env data-layout)
  (map (lambda (x) (cons (car x) (cddr x))) (data-tors data-layout)))

(define (dtor-name ctor-name index)
  (string->symbol
   (string-append (symbol->string ctor-name)
		  "~"
		  (number->string index))))

; for use in normalized form
(define lambda-arg caadr)
; for use elsewhere
(define lambda-args cadr)
(define lambda-body caddr)

(define (references prog)
  (ast-collect
   (lambda (x)
     (case (ast-type x)
       ['var (list x)]
       [else '()]))
   prog))

(define (graph bs)
  (define (go bs orig-bs)
    (if (null? bs)
	'(() . ())
	(let* [(bind (car bs))

	       (vert (car bind))
	       (refs (filter ; only count a reference if its a binding
		      (lambda (x) (assoc x orig-bs))
		      (references (cdr bind))))
	       (edges (map (lambda (x) (cons vert x))
			   refs))

	       (rest (if (null? (cdr bs))
			 (cons '() '())
			 (go (cdr bs) orig-bs)))
	       (total-verts (cons vert (car rest)))
	       (total-edges (append edges (cdr rest)))]
	  (cons total-verts total-edges))))
  (go bs bs))

(define (successors graph v)
  (define (go v E)
    (if (null? E)
	'()
	(if (eqv? v (caar E))
	    (cons (cdar E) (go v (cdr E)))
	    (go v (cdr E)))))
  (go v (cdr graph)))

					; takes in a graph (pair of vertices, edges)
					; returns a list of strongly connected components

					; ((x y w) . ((x . y) (x . w) (w . x))

					; =>
					; .->x->y
					; |  |
					; |  v
					; .--w

					; ((x w) (y))

					; this uses tarjan's algorithm, to get reverse
					; topological sorting for free
(define (sccs graph)
  
  (let* ([indices (make-hash-table)]
	 [lowlinks (make-hash-table)]
	 [on-stack (make-hash-table)]
	 [current 0]
	 [stack '()]
	 [result '()])

    (define (index v)
      (get-hash-table indices v #f))
    (define (lowlink v)
      (get-hash-table lowlinks v #f))

    (letrec
	([strong-connect
	  (lambda (v)
	    (begin
	      (put-hash-table! indices v current)
	      (put-hash-table! lowlinks v current)
	      (set! current (+ current 1))
	      (push! stack v)
	      (put-hash-table! on-stack v #t)

	      (for-each
	       (lambda (w)
		 (if (not (hashtable-contains? indices w))
					; successor w has not been visited, recurse
		     (begin
		       (strong-connect w)
		       (put-hash-table! lowlinks
					v
					(min (lowlink v) (lowlink w))))
					; successor w has been visited
		     (when (get-hash-table on-stack w #f)
		       (put-hash-table! lowlinks v (min (lowlink v) (index w))))))
	       (successors graph v))

	      (when (= (index v) (lowlink v))
		(let ([scc
		       (let new-scc ()
			 (let ([w (pop! stack)])
			   (put-hash-table! on-stack w #f)
			   (if (eqv? w v)
			       (list w)
			       (cons w (new-scc)))))])
		  (set! result (cons scc result))))))])
      (for-each
       (lambda (v)
	 (when (not (hashtable-contains? indices v)) ; v.index == -1
	   (strong-connect v)))
       (car graph)))
    result))


					; utils

(define (range s n)
  (if (= 0 n) '()
      (append (range s (- n 1))
	      (list (+ s (- n 1))))))

(define (flat-map f . xs) (fold-left append '() (apply map (cons f xs))))
(define (repeat x n) (if (<= n 0) '()
			 (cons x (repeat x (- n 1)))))


(define-syntax push!
  (syntax-rules ()
    ((_ s x) (set! s (cons x s)))))

(define-syntax pop!
  (syntax-rules ()
    ((_ s) (let ([x (car s)])
	     (set! s (cdr s))
	     x))))
